#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Box.H>
//#include <FL/Fl_Menu_Button.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Choice.H>
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Secret_Input.H>
#include <FL/Fl_Check_Button.H>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#ifdef WIN32
#include <io.h>
#endif

static char *g_nntp_listen_port;
static char *g_nntp_listen_address;
static char *g_bincache_listen_port;
static char *g_bincache_listen_address;

struct server_data {
        char *name;
        char *protocol;
        char *port;
        /* 11990 - Cacheplex default
         * 119 - NNTP default */
        char *fullfeed; /* 1 on NNTP */
        char *trace; /* Store. */
        char *cert; /* Store. */
        char *user;
        char *password;
};

static struct server_data *cacheplex_server;
static struct server_data **servers;
static struct server_data *g_current_server;
static int num_servers;
static Fl_Choice *g_serverlist_menu;
static Fl_Secret_Input *g_current_server_password;
static Fl_Input *g_current_server_address;
static Fl_Int_Input *g_current_server_port;
static Fl_Input *g_current_server_user;

void close_window(Fl_Widget *butt,void *window)
{
  Fl_Window *wind = (Fl_Window *)window;
  delete(wind);
}

void cb_nntplistenaddr_set(Fl_Widget *widget,void *data)
{
  if (g_nntp_listen_address) free(g_nntp_listen_address);
  g_nntp_listen_address = strdup((char *)data);
}

void redraw_serverspecific_settings(void)
{
  g_current_server_address->value(g_current_server->name);
  if (!g_current_server->port) {
    g_current_server->port = strdup("119");
  } else if (g_current_server->port[0]='\0') {
    free(g_current_server->port);
    g_current_server->port = strdup("119");
  }
  g_current_server_port->value(g_current_server->port);
  g_current_server_user->value(g_current_server->user);
  g_current_server_password->value(g_current_server->password);
  Fl::flush();
}

void cb_chooseserver(Fl_Widget *widget, void *server)
{
  g_current_server = (struct server_data *)server;
  redraw_serverspecific_settings();
}

void popup(const char *str)
{
  Fl_Window *window = new Fl_Window(200,100);
  Fl_Box *box = new Fl_Box(10,10,180,50,str);
  Fl_Button *ok = new Fl_Button(80,70,40,20,"OK");
  ok->callback(close_window,window);
  window->end();
  window->show();
}

void config_file_too_complex(void)
{
  Fl_Window *window = new Fl_Window(300,180);
  Fl_Box *box = new Fl_Box(10,10,280,160,
    "Your configuration file is too\n"
    "complicated for the friendly \n"
    "configuration app.  To edit it,\n"
    "either use a text editor or delete\n"
    "and recreate it.\n");
  Fl_Button *ok = new Fl_Button(130,150,40,20,"OK");
  ok->callback(close_window,window);
  window->end();
  char *x[2];
  x[0] = "iniedit";
  x[1] = NULL;
  window->show(1,x);
}

#ifdef WIN32
char *rindex(char *s, char c)
{
  int x = 0;
  char *last_c = NULL;
  while (s[x]) {
    if (s[x]==c) last_c = &s[x];
    x++;
  }
  return last_c;
}

char *index(char *s, char c)
{
  char *s2 = s;
  while (*s2) {
    if (*s2==c) return s2;
    s2++;
  }
  return NULL;
}

#endif

int parse_ini(void)
{
  FILE *f = fopen("bincache.ini","r");
  int ret = 1;
  static struct server_data *current_server;
  g_bincache_listen_port = strdup("0");
  #ifndef WIN32
  g_nntp_listen_port = strdup("11199");
  #else
  g_nntp_listen_port = strdup("119");
  #endif
  g_nntp_listen_address = strdup("127.0.0.1");
  g_bincache_listen_address = strdup("0.0.0.0");
  g_bincache_listen_port = strdup("0");
  if (!f) {
    servers = (struct server_data **)(calloc(1,sizeof(struct server_data *)));
    num_servers = 1;
    servers[0] = (struct server_data *)(calloc(1,sizeof(struct server_data)));
    servers[0]->name = strdup("");
    servers[0]->protocol = strdup("nntp");
    servers[0]->port = strdup("119");
    servers[0]->cert = strdup("");
    servers[0]->user = strdup("");
    servers[0]->password = strdup("");

    cacheplex_server = (struct server_data *)calloc(1,sizeof(struct server_data));
    cacheplex_server->protocol = strdup("cacheplex");
    cacheplex_server->name = strdup("cacheplex.deekoo.net");
    cacheplex_server->port = strdup("11990");
    cacheplex_server->cert = strdup("cacheplex.deekoo.net.cert.pem");
  } else {
    char buf[8192];
    char *str;
    while ((str=fgets(&buf[0],8192,f))) {
      char *newline = rindex(&buf[0],'\012');
      if (newline) {
        *newline = '\0';
      }
      newline = rindex(&buf[0],'\015');
      if (newline) {
        *newline = '\0';
      }
      if (buf[0]=='[') {
        if (!memcmp(&buf[1],"default]",8)) {
          config_file_too_complex();
          fclose(f);
          return 0;
        } else {
          if (current_server) {
            if (!strcmp(current_server->protocol,"cacheplex")) {
              if (cacheplex_server) {
                config_file_too_complex();
                fclose(f);
                return 0;
              } else {
                cacheplex_server = current_server;
              }
            } else {
              num_servers++;
              servers = (struct server_data **)realloc(servers,sizeof(struct server_data *)*num_servers);
              servers[num_servers-1] = current_server;
            }
          }
          char *rbracket = rindex(&buf[0],']');
          *rbracket = '\0';
          current_server = (struct server_data *)calloc(1,sizeof(struct server_data));
          current_server->name = strdup(&buf[1]);
          current_server->protocol = strdup("nntp");
        }
      } else if (!strncmp(&buf[0],"bincache-listen-port=",21)) {
        g_bincache_listen_port = strdup(&buf[21]);
      } else if (!strncmp(&buf[0],"cert=",5)) {
        current_server->cert = strdup(&buf[5]);
      } else if (!strncmp(&buf[0],"fullfeed=",9)) {
        current_server->fullfeed = strdup(&buf[9]);
      } else if (!strncmp(&buf[0],"nntp-listen-port=",17)) {
        g_nntp_listen_port = strdup(&buf[17]);
      } else if (!strncmp(&buf[0],"nntp-listen-address=",20)) {
        g_nntp_listen_address = strdup(&buf[20]);
      } else if (!strncmp(&buf[0],"password=",9)) {
        current_server->password = strdup(&buf[9]);
      } else if (!strncmp(&buf[0],"port=",5)) {
        current_server->port = strdup(&buf[5]);
      } else if (!strncmp(&buf[0],"protocol=",9)) {
        current_server->protocol = strdup(&buf[9]);
      } else if (!strncmp(&buf[0],"trace=",6)) {
        current_server->trace = strdup(&buf[6]);
      } else if (!strncmp(&buf[0],"user=",5)) {
        current_server->user = strdup(&buf[5]);
      } else if (buf[0]!='#' && buf[0]!='\0' && buf[0]!='\015') {
        fprintf(stderr,"Error: %s\n",&buf[0]);
        config_file_too_complex();
        fclose(f);
        return 0;
      }
    }
    num_servers++;
    servers = (struct server_data **)realloc(servers,sizeof(struct server_data *)*num_servers);
    servers[num_servers-1] = current_server;
  }
  if (!ret) {
    config_file_too_complex();
  }
  return ret;
}

void init_serverlist_menu(void)
{
  Fl_Menu_Item *serverlist = (Fl_Menu_Item *)calloc(num_servers+1,sizeof(Fl_Menu_Item));
  int x = 0;
  int menunum = 0;
  while (x<num_servers) {
    if (servers[x]==g_current_server) menunum = x;
    char *name = servers[x]->name;
    if ((!name) || (name[0]=='\0')) {
      name = strdup("New Server");
    }
    serverlist[x] = (Fl_Menu_Item){ name, 0, cb_chooseserver, (void *)servers[x] };
    x++;
  }
  serverlist[x] = (Fl_Menu_Item){0};
  g_serverlist_menu->menu(serverlist);
  g_serverlist_menu->value(menunum);
  g_serverlist_menu->redraw();
}

void cb_add_server(Fl_Widget *widget,void *data)
{
  num_servers+=1;
  servers = (struct server_data **)realloc(servers,num_servers*sizeof(struct server_data *));
  g_current_server = (struct server_data *)calloc(1,sizeof(struct server_data));
  g_current_server->name = strdup("");
  g_current_server->user = strdup("");
  g_current_server->protocol = strdup("nntp");
  g_current_server->port = strdup("119");
  g_current_server->trace = strdup("");
  g_current_server->fullfeed = strdup("1");
  g_current_server->cert = strdup("");
  servers[num_servers-1] = g_current_server;
  init_serverlist_menu();
  g_serverlist_menu->value(num_servers-1);
  redraw_serverspecific_settings();
}

void cb_del_current_server(Fl_Widget *widget, void *data)
{
  if (g_current_server->name) free(g_current_server->name);
  if (g_current_server->trace) free(g_current_server->trace);
  if (g_current_server->user) free(g_current_server->user);
  if (g_current_server->port) free(g_current_server->port);
  if (g_current_server->protocol) free(g_current_server->protocol);
  if (g_current_server->cert) free(g_current_server->cert);
  if (g_current_server->password) free(g_current_server->password);
  if (g_current_server->fullfeed) free(g_current_server->fullfeed);
  int menunum;
  if (g_current_server==servers[num_servers-1]) {
    if (num_servers>1) {
      free(g_current_server);
      num_servers -= 1;
      g_current_server = servers[num_servers-1];
      menunum = num_servers-1;
    } else {
      g_current_server->name = strdup("");
      g_current_server->port = strdup("119");
      g_current_server->trace = NULL;;
      g_current_server->cert = NULL;
      g_current_server->user = NULL;
      g_current_server->password = NULL;
      g_current_server->fullfeed = NULL;
      menunum = 0;
    }
  } else {
    free(g_current_server);
    struct server_data *new_server;
    int x = 0;
    menunum = 0;
    while (x<num_servers) {
      if (servers[x]==g_current_server) {
        menunum = x;
        new_server = servers[x+1];
        while (x+1>num_servers) {
          servers[x] = servers[x+1];
          x++;
        }
      }
      x++;
    }
    num_servers--;
    g_current_server = new_server;
  }
  init_serverlist_menu();
  g_serverlist_menu->value(menunum);
  redraw_serverspecific_settings();
}

void fprint_serverdata(FILE *f, struct server_data *server)
{
  fprintf(f,"[%s]\n",server->name);
  if (server->protocol && strcmp(server->protocol,"nntp")) {
    fprintf(f,"protocol=%s\n",server->protocol);
    if (!strcmp(server->protocol,"cacheplex")) {
      if (server->port && server->port[0] && strcmp(server->port,"11990")) {
        fprintf(f,"port=%s\n",server->port);
      }
    }
  } else {
    if (server->port && server->port[0] && strcmp(server->port,"119")) {
      fprintf(f,"port=%s\n",server->port);
    }
    if (server->fullfeed and strcmp(server->fullfeed,"1")) {
      fprintf(f,"fullfeed=%s\n",server->fullfeed);
    } else {
      fprintf(f,"fullfeed=1\n");
    }
  }
  if (server->trace && server->trace[0]) {
    fprintf(f,"trace=%s\n",server->trace);
  }
  if (server->cert && server->cert[0]) {
    fprintf(f,"cert=%s\n",server->cert);
  }
  if (server->user && server->user[0]) {
    fprintf(f,"user=%s\n",server->user);
  }
  if (server->password && server->password[0]) {
    fprintf(f,"password=%s\n",server->password);
  }
}

void cb_server_address(Fl_Widget *widget, void *data)
{
  if (g_current_server->name) {
    free(g_current_server->name);
  }
  Fl_Input *input = (Fl_Input *)widget;
  g_current_server->name = strdup(input->value());
  init_serverlist_menu();
}

void cb_server_password(Fl_Widget *widget, void *data)
{
  if (g_current_server->password) {
    free(g_current_server->password);
  }
  Fl_Input *input = (Fl_Input *)widget;
  g_current_server->password = strdup(input->value());
}

void cb_server_user(Fl_Widget *widget, void *data)
{
  if (g_current_server->user) {
    free(g_current_server->user);
  }
  Fl_Input *input = (Fl_Input *)widget;
  g_current_server->user = strdup(input->value());
}

void set_input_tcp_port(Fl_Input *input, char **portvar,int minval)
{
  int i = atoi(input->value());
  if (i<minval || i>65535) {
    if (*portvar) {
      input->value(*portvar);
    }
    input->redraw();
    return;
  }
  if (i==0) {
    input->value("0");
    input->redraw();
  }
  if (*portvar) {
    free(*portvar);
  }
  *portvar = strdup(input->value());
}

void cb_server_port(Fl_Widget *widget, void *data)
{
  set_input_tcp_port((Fl_Input *)widget,&g_current_server->port,1);
}

void cb_nntp_listen_port(Fl_Widget *widget, void *data)
{
  set_input_tcp_port((Fl_Input *)widget,&g_nntp_listen_port,0);
}

void cb_bincache_listen_port(Fl_Widget *widget, void *data)
{
  set_input_tcp_port((Fl_Input *)widget,&g_bincache_listen_port,0);
}

void cb_save_settings(Fl_Widget *widget, void *data)
{
  int x = 0;
  while (x<num_servers) {
    if ((!servers[x]->name) 
      || (servers[x]->name[0]=='\0') || index(servers[x]->name,']')) {
      popup("All servers must be named.");
      g_serverlist_menu->value(x);
      return;
    }
    x++;
  }
  x = 0;
  FILE *newcf = fopen("bincache.nni","wb");
  fprintf(newcf,"nntp-listen-port=%s\n",g_nntp_listen_port);
  fprintf(newcf,"nntp-listen-address=%s\n",g_nntp_listen_address);
  fprintf(newcf,"bincache-listen-port=%s\n",g_bincache_listen_port);
  fprint_serverdata(newcf,cacheplex_server);
  while (x<num_servers) {
    fprint_serverdata(newcf,servers[x]);
    x++;
  }
  fclose(newcf);
  #ifdef WIN32
  unlink("bincache.bni");
  #endif
  rename("bincache.ini","bincache.bni");
  #ifdef WIN32
  unlink("bincache.ini");
  #endif
  rename("bincache.nni","bincache.ini");
  delete((Fl_Window *)data);
}

int main(int argc, char **argv)
{
  if (parse_ini()) {
    Fl_Window *window = new Fl_Window(500,230);
    Fl_Choice *nntplistenaddr = new Fl_Choice(245,5,90,20,"Listen for NNTP connections from ");
    Fl_Menu_Item nntplistenaddr_menu[4] = {
      { "localhost", 0, cb_nntplistenaddr_set, (void *)"127.0.0.1" },
      { "anywhere", 0, cb_nntplistenaddr_set, (void *)"0.0.0.0" },
      { 0 },
      { 0 }
      };
    nntplistenaddr->menu(nntplistenaddr_menu);
    if (!strcmp(g_nntp_listen_address,"0.0.0.0")) {
      nntplistenaddr->value(1);
    } else if (strcmp(g_nntp_listen_address,"127.0.0.1")) {
      nntplistenaddr_menu[2] = (Fl_Menu_Item){ g_nntp_listen_address, 0, cb_nntplistenaddr_set, (void *)g_nntp_listen_address };
      nntplistenaddr->menu(nntplistenaddr_menu);
      nntplistenaddr->value(2);
    }
    Fl_Int_Input *nntpport = new Fl_Int_Input(255,25,60,20,"NNTP port (0 to pick a random port)");
    nntpport->value(g_nntp_listen_port);
    nntpport->callback(cb_nntp_listen_port);
    Fl_Int_Input *bincacheport = new Fl_Int_Input(358,45,60,20,"Bincache peer port (0 for automatic port selection)");
    bincacheport->callback(cb_bincache_listen_port);
    bincacheport->value(g_bincache_listen_port);
    Fl_Check_Button *g_use_premium_cacheplex = new Fl_Check_Button(5,65,20,20,"Use premium Cacheplex server");
    Fl_Input *g_cacheplex_premium_key = new Fl_Input(117,85,320,20,"Cacheplex key: ");
    g_serverlist_menu = new Fl_Choice(80,115,200,20,"Server ");
    init_serverlist_menu();
    Fl_Button *add_server = new Fl_Button(295,115,90,20,"New server");
    add_server->callback(cb_add_server);
    Fl_Button *del_server = new Fl_Button(395,115,100,20,"Delete server");
    del_server->callback(cb_del_current_server);
    g_current_server = servers[0];
    g_current_server_address = new Fl_Input(75,135,300,20,"Address: ");
    g_current_server_address->callback(cb_server_address);
    g_current_server_port = new Fl_Int_Input(425,135,60,20,"Port: ");
    g_current_server_port->callback(cb_server_port);
    g_current_server_user = new Fl_Input(89,155,300,20,"Username: ");
    g_current_server_user->callback(cb_server_user);
    g_current_server_password = new Fl_Secret_Input(85,175,300,20,"Password: ");
    g_current_server_password->callback(cb_server_password);
    Fl_Button *save_settings = new Fl_Button(40,200,160,20,"Save settings");
    save_settings->callback(cb_save_settings, window);
    Fl_Button *cancel = new Fl_Button(280,200,160,20,"Cancel");
    cancel->callback(close_window,window);
    redraw_serverspecific_settings();
    window->end();
    window->show(argc,argv);
  }
  return Fl::run();
}
